""":class:`~apache_beam.transforms.ptransform.PTransform` is for reading from
and writing to relational databases. SQLAlchemy_ is used for interfacing the
databases.
.. _SQLAlchemy: https://www.sqlalchemy.org/
"""

from __future__ import division, print_function

from apache_beam import PTransform, Create, ParDo, DoFn

from beam_bc365.io.bc_service_api import (
    ServiceConfiguration,
    BusinessCentralSource
)

# It is intended ServiceConfiguration is imported
# from this module by the library users. Below is to make sure they've been
# imported to this module.
assert ServiceConfiguration is not None

class ReadFromService(PTransform):
    """A :class:`~apache_beam.transforms.ptransform.PTransform` for reading
    tables on relational databases.
    It outputs a :class:`~apache_beam.pvalue.PCollection` of
    ``dict:s``, each corresponding to a row in the target database table.
    Args:
        source_config (SourceConfiguration): specifies the target database.
    Examples:
        Reading from a web service on Microsoft Central Business Central  ::
            import apache_beam as beam
            from apache_beam.options.pipeline_options import PipelineOptions
            from beam_bc365.io import bc_service
            service_config = bc_service.ServiceConfiguration(
                username='foo',
                service_key='api-key',
                companies=['Cranos'],
                service='G_L_ENTRY',
                instance='Sandbox',
                instance_id="2E0B815D-A1AC-491B-BD09-0876DACC2A12"
            )
            with beam.Pipeline(options=PipelineOptions()) as p:
                records = p | "Reading records from service" >>  bc_service.ReadFromService(
                    service_config=service_config,
                )
                records | 'Writing to stdout' >> beam.Map(print)
        The output will be something like ::
            {'name': 'Jan', 'num': 1}
            {'name': 'Feb', 'num': 2}
        Where "name" and "num" are the column names.
    """

    def __init__(self, service_config, *args, **kwargs):
        super(ReadFromService, self).__init__(*args, **kwargs)
        self._read_args = dict(service_config=service_config)

    def expand(self, pcoll):
        return (
            pcoll
            | Create([self._read_args])
            | ParDo(_ReadFromWebService())
        )


class _ReadFromWebService(DoFn):
    def process(self, element):
        bc_args = dict(element)
        source = BusinessCentralSource(**bc_args)
        try:
            for record in source.read_data():
                yield record
        except:
            raise




from __future__ import division, print_function

import urllib
import requests

from requests.auth import HTTPBasicAuth


class ServiceConfiguration():
    """Holds parameters for accessing a Microsfot Business Central 365.
    User to pass database access parameters to
    :class:`~beam_nuggets.io.relational_db_api.SqlAlchemyDB`.
    ``SourceConfiguration.url`` provides the database url used by SqlAlchemy to
    connect to the database.
    Args:
        username (str): Username for Business Central 365.
        service_key (str): API Key for Business Central 365.
        companies (list): List of Companies to call endpoints for.
        service (str): The service to call.
        instance (str): The Business Central 365 Instance.
        instance_id (str): The Ownership instance if
            found in the Business Central 365 URL.
    """
    def __init__(
        self,
        username,
        service_key,
        companies,
        service,
        instance,
        instance_id,
        base_url="https://api.businesscentral.dynamics.com/v2.0"
    ):
        self.url_data = self._get_urls(
            service=service,
            companies=companies,
            instance=instance,
            instance_id=instance_id,
            base_url=base_url
        )
        self.auth = HTTPBasicAuth(username, service_key)

    @staticmethod
    def _get_urls(service,companies,instance,instance_id,base_url):
        url_data = []
        for company in companies:
            url_data.append({
               'endpoint': f"{base_url}/{instance_id}/{instance}/ODataV4/Company('{urllib.parse.quote(company)}')/{service}",
               'company': company
            })
        return url_data


class BusinessCentralSource(object):

    def __init__(self, service_config):
        self._service_config = service_config

    def read_data(self):
        for endpoint_data in self._service_config.url_data:
            for record in self.service_data(endpoint_data):
                yield record
    
    def service_data(self, endpoint_data):
        headers = {
            "OData-MaxVersion": "4.0",
            "OData-Version": "4.0",
            "Accept": "application/json",
            "Content-Type": "application/json; charset=utf-8",
            "Prefer": "odata.include-annotations=OData.Community.Display.V1.FormattedValue"
        }
        res = requests.get(
            endpoint_data.get('endpoint'),
            auth=self._service_config.auth,
            headers=headers
        )
        for record in res.json().get('value'):
            # Company gets added to records for instances that collect multiple companies
            record['Company'] =  endpoint_data.get('company')
            yield record



from __future__ import division, print_function

from setuptools import setup, find_packages


REQUIRED_PACKAGES = [
    'apache-beam>=2.8.0,<3.0.0',
    'requests'
]

REQUIRED_PACKAGES_TEST = [
    'nose>=1.3.7,<2.0.0',
    'pylint',
    'responses'
]

REQUIRED_PACKAGES_CI_CD = [
    "bump2version",
    "twine"
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='beam-bc365',
    version='0.2.0',
    author='Brendan Kamp',
    author_email='brendankamp757@gmail.com',
    description='Microsoft Business Central transofrm for the Apache beam python SDK.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Spazzy757/beam-bc365",
    install_requires=REQUIRED_PACKAGES,
    extras_require={'dev': REQUIRED_PACKAGES_TEST + REQUIRED_PACKAGES_CI_CD},
    packages=find_packages(exclude=("test", "tests")),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
